#!/bin/bash
# ======================================================================
#
# ICINGA PASSIVE CLIENT
#
# Run all passive checks and send response to Icinga Endpoint
#
# Requirements
# - curl
# - jq
# ----------------------------------------------------------------------
# ah = axel.hahn@unibe.ch
# 2021-03-..  init
# 2022-01-11  v0.7   ah  shellcheck
# 2022-02-16  v0.8   ah  add --cfg param
# 2022-03-04  v0.9   ah  abort on http 5xx error
# 2022-03-14  v0.10  ah  less output and add _elog to run as a service
# 2022-07-08  v0.11  ah  check pipes in output and performance data
# 2023-01-30  v0.12  ah  remove pipes with sed remove the pipe before performancedata
# 2023-02-13  v0.13  ah  rename myHost to MY_NAME (like in director_cli)
# 2023-02-17  v0.14  ah  add hostnamme into CFGSTORAGE
# 2023-02-17  v0.15  ah  check inc_getconfig.sh exists and hostname -f has a value
# 2023-11-01  v0.16  ah  generate json with "jo"
# 2023-11-02  v0.17  ah  fix line breaks of sent output, shell check fixes
# 2023-11-02  v0.18  ah  fix quoting of perfdata in json
# 2023-11-02  v0.19  ah  move _parseConfig
# 2024-11-20  v0.20  ah  update rest api client; use cert "${dir_cfg}/certs/ca.crt"
# ======================================================================


_product="ICINGA PASSIVE CLIENT"
_version="0.20"
_license="GNU GPL 3.0"
_copyright='(c) Institute for Medical Education * University of Bern'

typeset -i debug=0

which curl >/dev/null || exit 1
which jo   >/dev/null || exit 1

# source config ...
_cfg="$( dirname $0 )/inc_getconfig.sh"
if [ ! -f "$_cfg"  ]; then
  echo "ERROR: The configuration file $_cfg was not created yet."
  exit 1
fi
. "$_cfg"

logfile="${dir_logs}/execution.log"

MY_NAME=$(hostname -f)
if [ -z "$MY_NAME" ]; then
  echo "ERROR: hostname is empty. Ensure that the command 'hostname -f' returns a fqdn."
  exit 1
fi

# for loop mode only: max. random sleep time
typeset -i sleeptime=30

typeset -i _rc_all=0

# ----------------------------------------------------------------------
#
# FUNCTIONS
#
# ----------------------------------------------------------------------
# ..................................................................
#
# helper to make http base setup for host actions
function _initHttp(){
  # see inc_functions
   _initHttpWithConfigfile "${dir_cfg}/api-icinga2.cfg"
  if [ $debug -ne 0 ]; then
    http.setDebug 1
  fi
  local _cert="${dir_cfg}/certs/ca.crt"
  if [ ! -f "$_cert" ]; then
    echo
    echo "ERROR: certificate was not found."
    echo "Copy the Icinga master CA into $_cert"
    exit 1
  fi
  http.setCA "$_cert"
}


# ......................................................................
#
# find first place of the check script in the known plugin dirs 
# see ${dir_plugins} in checks.cfg
# param  string  name of the check script without path
#
function findCheckScript(){
	local _script=$1
	for mydir in ${dir_plugins}
	do
		if [ -x "${mydir}/${_script}" ]; then
			echo "${mydir}/${_script}"
		fi
	done | head -1
}

# helper used function in loopChecks
# get a snapshot of a few files
function _getFileSnapshot(){
  ls -l $(dirname $0)/* ${dir_cfg}/*
}


# ......................................................................
#
# Loop over executing all checks
# no params
#
function loopChecks(){

      local lockfile
      lockfile="${dir_data}/loop.pid"
      local snapShotStart
      snapShotStart=${dir_data}/$(basename $0)-start.fingerprint
      local snapShotCurrent
      snapShotCurrent=${dir_data}/$(basename $0)-last.fingerprint
      if [ -f "${lockfile}" ]; then
        local lockpid
        lockpid=$(cat "${lockfile}" | cut -f 2 -d "-" | cut -f 4 -d " " | grep "[0-9]")
        ps -f --pid "$lockpid" | grep "$(basename $0)" | grep loop >/dev/null
        if [ $? -eq 0 ]; then
          _elog "ABORT: Loop seems to run already. See process with PID $lockpid"
          _elog $( ps -f --pid "$lockpid" )
          exit 0
        fi
      fi

      _log "---------- starting in a permanent loop"
      icingaHostMustExist

      echo "Serviceloop started $(date) - process id $$" > "${lockfile}"
      if [ $? -ne 0 ]; then
          _elog "ABORT: Lock file is not writable ${lockfile}."
          _elog $( ls -l "${lockfile}" );
          exit 1
      fi

      _getFileSnapshot>"${snapShotStart}"
      if [ $? -ne 0 ]; then
          _elog "ABORT: Snapshot file is not writable ${snapShotStart}."
          _elog $( ls -l "${snapShotStart}" )
          exit 1
      fi
      while true; do
        # typeset -i local iSleep=$(($RANDOM%$sleeptime))
        # sleep minimum is half of $sleeptime
        local iSleep
        typeset -i iSleep=$(($RANDOM%$sleeptime/2+$sleeptime/2))
        _log "sleeping $iSleep sec ..."
        sleep $iSleep
        _log "______________________________________________________________________"
        _log ""
        if ! _getFileSnapshot > "$snapShotCurrent"; then
            _elog "ABORT: Snapshot file is not writable ${snapShotCurrent}."
            _elog $( ls -l "${snapShotCurrent}" )
            exit 1
        fi
        if ! diff "$snapShotStart" "$snapShotCurrent" >/dev/null; then
          _elog "ABORT: Files were updated / overwritten. The loop must be restarted.\n$(diff  $snapShotStart $snapShotCurrent)"
          exit 1
        fi
        processAllChecks
      done
}
# ......................................................................
#
# execute all defined checks one by one
# no params
#
function processAllChecks(){
  # loop over all defined checks
  local iChecksTotal; typeset -i iChecksTotal
  local iCounter;     typeset -i iCounter
  local iLoopStart;   typeset -i iLoopStart
  local iLoopEnd;     typeset -i iLoopEnd
  local iLoopTime;    typeset -i iLoopTime

  iChecksTotal=$(getChecks | wc -l)
  iCounter=0
  _rc_all=0
  iLoopStart=$(_getUnixTs)

  _log ""
  _log "------ looping over all checks"
  for myconfig in $(getChecks)
  do
    iCounter=$iCounter+1
    _log "--- processing [$iCounter of $iChecksTotal] $myconfig"
    processCheck "$myconfig"
    _log ""

  done
  iLoopEnd=$(_getUnixTs)
  iLoopTime=$iLoopEnd-$iLoopStart

  _log "------ loop done - needed $iLoopTime sec - rc=$_rc_all"
}

# actions for icinga host
# param  string  action; "get" only
function icingaHost(){
  local _logPrefix="${MY_NAME} :: API |"
  local _apiRequest=objects/hosts/${MY_NAME}
  local _localCache=${dir_data}/host_${MY_NAME}_deployed-at-icinga.txt
  local _iRefreshCache; typeset -i _iRefreshCache=120

  local sAction=$1

  _initHttp

  case $sAction in
    'get')
      # update after caching was added in http-component
      # http.responseImport $_localCache
      # typeset -i local iAgeLastGet=`http.getRequestAge`
      # _log "${_logPrefix} INFO: cache is $iAgeLastGet sec old ... TTL is _iRefreshCache=3600 sec"
      # if [ $iAgeLastGet -eq 0 -o $iAgeLastGet -gt $_iRefreshCache ]; then
      #   _log "${_logPrefix} INFO: request to Icinga GET $_apiRequest"
      #   _getApiObject $_apiRequest $_localCache
      # fi

      # new:
      http.setCacheTtl     $_iRefreshCache
      http.setCacheFile    $_localCache
      http.makeRequest GET $_apiRequest
      if http.isServerError >/dev/null; then
        _elog "CRITICAL ERROR: Icinga2 API request failed with a server error GET $_apiRequest"
        exit 1
      fi

      # set return code of GET action
      http.isOk >/dev/null
      ;;
    *)
      _log "ERROR: unknown action parameter $sAction"
  esac

}

# for check on the beginning of the script:
# execute a check only if the host exists on icinga2
# global  string  MY_NAME 
# global  string  dir_data
function icingaHostMustExist(){
  _log "check if the host [${MY_NAME}] exists on Icinga ..."
  icingaHost get
  if [ $? -ne 0 ]; then
    _echo $( http.getResponse )
    if [ "$(http.getStatuscode)" = "000" ]; then
      _elog "ERROR: Unable to reach the Icinga node. Stopping script current monitoring actions."
      exit 1
    fi
    _elog "ERROR: host object for ${MY_NAME} is not available on Icinga service (yet) - Status: $(http.getStatuscode)"
    _echo
    _echo "ABORTING"
    _echo
    _echo "To run checks ..."
    _echo "- you must create the host on director (check director-cli.sh --hr)"
    _echo "- the director must deploy the host to icinga daemon"
    _echo
    rm -f "${dir_data}"/service__check* 2>/dev/null
    exit 1
  fi
  _log "OK, found."
}


# ......................................................................
#
# process a single check
# param  string  name of config file
# param  string  name of config file
#
function processCheck(){
  local _myconfig=$1
  local _force=$2

  _parseCheckConfig "${_myconfig}"

  local iPipes;      typeset -i iPipes
  local iCheckStart; typeset -i iCheckStart

  local _outfile=${dir_data}/service__check__${checkName}__output.txt
  local _output
  local _FoundErrors
  local _response=${dir_data}/service__check__${checkName}__icinga_response.txt
  local _rc; typeset -i _rc=0

  local iTsEnd;   typeset -i iTsEnd
  local iTsStart; typeset -i iTsStart

  iCheckStart=$(_getUnixTs)

  local _logPrefix="${checkName} |"
  _log "${_logPrefix} INFO: every ${checkInterval} sec: ${checkCommand}"

  _initHttp

  # --- check last run ... if never or > $interval then execute
  doRun=0
  if [ ! -f "$_outfile" ]; then
    _log "${_logPrefix} INFO: Never executed before"
    doRun=1
  else
    typeset -i iAgeLastRun
    iAgeLastRun=$(_getFileAge "$_outfile")
    _log "${_logPrefix} INFO: last run was $iAgeLastRun sec ago ... vs Interval = $checkInterval ... sleeptime = $sleeptime"
    iAgeLastRun=$iAgeLastRun+$sleeptime
    if [ $iAgeLastRun -gt $checkInterval ]; then
      doRun=1
    fi

    if [ ! -z "$_force" ]; then
      doRun=1
      _log "${_logPrefix} INFO: forced execution by given param "
    fi
  fi

  if [ $doRun -ne 0 ]; then
    myscript=$(echo "$checkCommand" | cut -f 1 -d " ")
    myFullscript=$(findCheckScript "$myscript")
    if [ -z "$myFullscript" ]; then
      _log "${_logPrefix} ERROR: $myscript was not found in any plugin dir"
    else
      myparams=$( echo $checkCommand | grep " " | cut -f 2- -d " " )

      #
      # --- this executes the check plugin ...
      #
      _log "${_logPrefix} starting $myFullscript $myparams"
      iTsStart=$(date +%s)
      eval $myFullscript $myparams > "$_outfile"
      rc=$?
      if [ ! -w $_outfile ]; then
            _elog "${_logPrefix} ERROR: output file $_outfile is not writable."
            _elog "${_logPrefix} $( ls -ld ${dir_data} $_outfile )"
            exit 1
      fi
      test $rc -gt 0 && _FoundErrors+="Exitcode of check is greater zero ... "
      iTsEnd=$(date +%s)
      outPerfdata=$(grep '|' $_outfile | rev | cut -f 1 -d '|' | rev | grep "=")
      _output="$( cat "$_outfile" )"
      _echo
      _echo -------- check output:
      _echo "$_output"

      iPipes=$( grep -o "|" < "$_outfile" | wc -l )
      # _echo "Pipe chars: $iPipes"
      if [[ $iPipes -gt 1 ]]; then
        _elog "WARNING - pipes were found in plugin output"
        _elog "command was: $myFullscript $myparams"
        _log "${_logPrefix} WARNING - pipes were found in plugin output"
      fi

      _echo

      #
      # --- extract performance data:
      #
      # if ! test -n "$outPerfdata" && echo "$outPerfdata" | grep "=[0-9\.]*;[0-9\.]*;[0-9\.]*;[0-9\.]*;" >/dev/null; then
      #   _elog "WARNING - this does not look like performance data: $outPerfdata"
      #   _elog "command was: $myFullscript $myparams"
      #   _log "${_logPrefix} this does not look like performance data: $outPerfdata"
      #   outPerfdata=""
      # fi
      _log "${_logPrefix} check command finished with returncode $rc"

      _rc=$_rc+$rc

      #
      # --- send check result to Icinga
      #     fields of the object
      #     https://icinga.com/docs/icinga2/latest/doc/12-icinga2-api/#process-check-result
      #
      data="$( jo  \
          check_source=${MY_NAME}                 \
          check_command="""${myFullscript} $myparams"""          \
          exit_status=$rc                         \
          ttl=$checkInterval                      \
          execution_start=$iTsStart               \
          execution_end=$iTsEnd                   \
          performance_data="${outPerfdata}"   \
          plugin_output="""${_output}""" 
      )"

      slot="$(_getName4Svcathost ${checkName} | sed 's# #%20#g')"

      _log "${_logPrefix} starting POST of data to monitoring server"
      _echo POST actions/process-check-result?service=${MY_NAME}!${slot} "$data"
      _APIcall POST actions/process-check-result?service=${MY_NAME}!${slot} "$data"
      http.responseExport "$_response"
      if [ ! -w "$_response" ]; then
            _elog "${_logPrefix} ERROR: responsefile $_response is not writable."
            _elog "${_logPrefix} $( ls -ld ${dir_data} $_response )"
            exit 1
      fi

      # --- check if data were sent successfully
      if http.isOk >/dev/null; then
        _log "${_logPrefix} rc=$rc - OK, response was sent to Icinga"
      else
        local httpcode; httpcode=$(http.getStatuscode)
        _rc=$_rc+1
        case $httpcode in
          404) _FoundErrors+="Icinga response $httpcode: host or service check was not created in Icinga yet ... " ;;
          *)   _FoundErrors+="Icinga response was not sent to Icinga: $httpcode ... " ;;
        esac
        _echo
        _elog "${_logPrefix} rc=$rc - $_FoundErrors"
        # _echo
        # _echo For Debugging:
        # _echo "$data"
        _log "$data"
      fi

    fi

  else
    _log "${_logPrefix} SKIP execution."
  fi

  # add current result to global returncode
  _rc_all=$_rc_all+$_rc

  local iCheckEnd; typeset -i iCheckEnd
  local iCheckTime; typeset -i iCheckTime

  iCheckEnd=$(_getUnixTs)
  iCheckTime=$iCheckEnd-$iCheckStart

  _log "${_logPrefix} finished after $iCheckTime sec with returncode $_rc"
}
# ----------------------------------------------------------------------
# help
# ----------------------------------------------------------------------

# show help text
function showHelp(){
self=$(basename $0)
cat <<EOH


INTRODUCTION

$_product v$_version

Handle and execute icinga passive checks.
With this client can run a single check, all checks or make a permanent loop.
A new local check will be added to Icinga while running it the first time.


GENERAL PARAMETERS

  --cfg CONFIGFILE
    load a costom config file; default: ./inc_getconfig.sh
    This must be the 1st parameter to be processed.

  --help or -h or -?
    show this help and abort.

  --version or -v
    show the version abd abort. It must be the 1st parameter.

SERVICE ACTIONS

  --list
    get a list of local config files

  --loop
    Start to check all passive checks in a permanent loop.
    It makes a random sleep if $sleeptime sec between all loops. This shuffles
    the access time of all clients making requests to the icinga server.

    Multiple starts will be detected. This parameter is the optimal choice for a
    cronjob.

  --runonce
    Start to check all passive checks once.
    This method respects the interval per check. Only outdated checks will be
    executet.
    This is a second choice for a cronjob if the runtime of all checks
    is much shorter than your cronjob interval to prevent multiple processes.
    Multiple starts will NOT be detected.

  --run CONFIGFILE
    Run a check by pointing the config file (see --list)
    This execution ignores the interval and forces the execution.


CONFIG

  Config file
  /etc/icinga2-passive-client/client.cfg
  It sets other used pathes.

  Config files for checks are in ${dir_cfg}/checks/


DEBUGGING

  The Output of check and results from Icinga are in
  ${dir_data}.

  A log of all performed executed check runs by $self
  are in $logfile.
  BTW: do not forget to add a log rotation for it.

EOH
}


# ----------------------------------------------------------------------
#
# MAIN
#
# ----------------------------------------------------------------------

. "$( dirname $0 )/inc_functions.sh"

# show version without header
if [ "$1" = "--version" ] || [ "$1" = "-v" ]; then
  echo "$( basename $0) v$_version"
  echo "$_license .. $_copyright"
  exit 0
fi


_echo "
______________________________________________________________________________________

 _______        __
|_     _|.----.|__|.-----.-----.---.-.
 _|   |_ |  __||  ||     |  _  |  _  |
|_______||____||__||__|__|___  |___._|
                         |_____|
 ______                     __                   ______ _____   __               __
|   __ \.---.-.-----.-----.|__|.--.--.-----.    |      |     |_|__|.-----.-----.|  |_
|    __/|  _  |__ --|__ --||  ||  |  |  -__|    |   ---|       |  ||  -__|     ||   _|
|___|   |___._|_____|_____||__| \___/|_____|    |______|_______|__||_____|__|__||____|
                                                                                 v${_version}

  $_license .. $_copyright

______________________________________________________________________________________

"

_elog "Starting $_product $_version"


if [ "$1" = "--cfg" ] && [ -n "$2" ]; then
  echo "INFO: loading custom config [$2]..."
  . "${2}"
  shift 2
else
  . "$( dirname $0 )/inc_getconfig.sh"
fi

. "$( dirname $0 )/inc/http.class.sh"

if [ $# -eq 0 ]; then
  showHelp
  exit 0
fi

if [ -z "${dir_cfg}" ]; then
  echo ERROR: $_product is not installed/ configured yet on this machine.
  exit 1
fi

touch ${logfile}


while [ $# -gt 0 ];
do
  case "$1" in

    '--help' | '-h' | '-?')
      showHelp
      exit 0
      ;;

    '--list')
      getChecks
      ;;

    '--loop')
      loopChecks
      ;;

    '--runonce')
      processAllChecks
      ;;

    '--run')
      processCheck "$2" "force"
      shift 1
      ;;

    *)
      echo "ERROR: unknown parameter detected."
      exit 2
  esac
  shift 1
done
echo

# remark:
# $_rc_all is a collected status code in the loop of all actions
# if it is 0 then all checks were OK and have been sent to Icinga
# echo exit with status code $_rc_all
# exit $_rc_all
exit 0

# ----------------------------------------------------------------------
